
import numpy as np
import scipy
import scipy.sparse as sp
from scipy.spatial.distance import pdist, cdist, squareform
from itertools import combinations
from scipy.sparse.csgraph import shortest_path

d = np.array([
		[0.,.11111111,9.,np.inf],
		[.11111111,0.,0.25,np.inf],
		[9.,0.25,0.,0.66666667],
		[np.inf,np.inf,0.66666667,0.],
], dtype=float)


ds = sp.csr_matrix(d)

apsp = shortest_path(d, method='auto', directed=False)

print d
print apsp


# -*- coding: utf-8 -*-
"""
Utility functions for the Distance Closure package
"""
#    Copyright (C) 2015 by
#    Luis Rocha <rocha@indiana.edu>
#    Thiago Simas <@.>
#    Rion Brattig Correia <rionbr@gmail.com>
#    All rights reserved.
#    MIT license.
import numpy as np

__author__ = """\n""".join(['Luis Rocha <rocha@indiana.com>',
							'Thiago Simas <@.>',
							'Rion Brattig Correia <rionbr@gmail.com>'])
__all__ = [
			'dist2prox',
			'prox2dist',
		]

__metrics__ = ['metric','ultrametric','semantic']


def prox2dist(O):
	"""
	Transforms a matrix of non-negative [0,1] proximities x to
	distance weights in the [0,inf] interval:
		      1
		s = ----- - 1
		      x
	Parameters
	----------
	x : array_like
		an array of non-negative distances.
	"""
	if (type(O).__module__ == 'numpy'):
		return prox2dist_numpy(O)
	elif (type(O).__module__.split('.')[1] == 'sparse'):
		return prox2dist_sparse(O)
	else:
		raise ("Format not accepted: try numpy or scipy.sparse formats")


def prox2dist_sparse(A):
	A.data = prox2dist_numpy(A.data)
	return A

def prox2dist_numpy(A):
	
	def _prox2dist(x):
		if x == 0:
			return np.inf
		else:
			return (1/float(x)) - 1

	f = np.vectorize(_prox2dist)
	return f(A)

def dist2prox(O):
	"""
	Transforms a matrix of non-negative integer distances x to
	proximity/similarity weights in the [0,1] interval:
		      1
		s = -----
		    x + 1
	Parameters
	----------
	x : array_like
		an array of non-negative distances.
	"""
	if (type(O).__module__ == 'numpy'):
		return dist2prox_numpy(O)
	elif (type(O).__module__.split('.')[1] == 'sparse'):
		return dist2prox_numpy(O)
	else:
		raise ("Format not accepted: try numpy or scipy.sparse formats")

def dist2prox_sparse(A):
	A.data = dist2prox_numpy(A.data)
	return A

def dist2prox_numpy(A):

	def _dist2prox(x):
		if x == np.inf:
			return 0
		else:
			return (x + 1) ** -1

	f = np.vectorize(_dist2prox)
	return f(A)





# -*- coding: utf-8 -*-
"""
Compute the Jaccard coefficient (distance/similarity) over a matrix

"""
#    Copyright (C) 2015 by
#    Rion Brattig Correia <rionbr@gmail.com>
#    All rights reserved.
#    MIT license.
import numpy as np
import scipy.sparse as sp
from scipy.sparse import csr_matrix, lil_matrix
from scipy.spatial.distance import cdist, squareform
from itertools import combinations
import warnings

__author__ = """\n""".join(['Rion Brattig Correia <rionbr@gmail.com>'])
__all__ = ['pairwise_proximity']
__metrics__ = [
				'jaccard','jb','jaccard_bitwise', # Bitwise Jaccard Coefficient
				'jaccard_set','js', # Set Comparison Jaccard Coefficient
				'jaccard_weighted','weighted_jaccard','wj', # Weighted Jaccard
				]

def pairwise_proximity(M, metric='jaccard', *args, **kwargs):
	"""
	Calculates pairwise proximity coefficient between rows of a matrix

	Parameters
	----------
	M : matrix, numpy or scipy

    metric : string
    	type of jaccard proximity to use.
    	- Binary comparison = 'jaccard','jb','jaccard_bitwise' =
    	- Set comparison ='jaccard_set','js'
    	- Weighted comparison = 'weighted_jaccard','wj'

    Note: Jaccard distance = ( 1 - Coefficient )

	Returns
	--------
	M : matrix
		The matrix of proximities
	
	Usage
	------
	P = pairwise_proximity(M, metric='jaccard')

	"""

	# Numpy object
	if (type(M).__module__ == np.__name__):
		return _pairwise_proximity_numpy(M, metric, *args, **kwargs)
	elif (sp.issparse(M)):
		return _pairwise_proximity_sparse(M, metric, *args, **kwargs)
	else:
		raise TypeError("Input is not a valid object, try a numpy array")


def _pairwise_proximity_numpy(M, metric='jaccard', *args, **kwargs):	
	""" Pairwise proximity computation over dense matrix (numpy) """
	coef = _get_dense_metric_function(metric)

	# If is not a Numpy array
	if type(M) != 'numpy.ndarray':
		M = np.array(M)

	# If matrix has negative entries
	if M.min() < 0:
		raise TypeError("Matrix cannot have negative numbers")

	# Calculate proximity
	m, n = M.shape
	dm = np.zeros((m * (m - 1)) // 2, dtype=np.double)
	
	k = 0
	for i in xrange(0, m - 1):
		print 'calc row:',i,'of',m-1
		for j in xrange(i + 1, m):
			dm[k] = coef(M[i,:], M[j,:], *args, **kwargs)
			k += 1
	
	dm = squareform(dm) #Make into a matrix format
	np.fill_diagonal(dm, 1) #Fill diagonal
	return dm

def _pairwise_proximity_sparse(X, metric='jaccard'):
	""" Pairwise proximity computation over sparse matrix (scipy.sparse) """
	how, coef = _get_sparse_metric_function(metric)
	
	if how == 'indices':
		generator = (coef(row1.indices, row2.indices) for row1, row2 in combinations(X, r=2))
	elif how == 'values':
		generator = (coef(row1.data, row2.data) for row1, row2 in combinations(X, r=2))
	elif how == 'both':
		generator = (coef(row1.toarray(), row2.toarray()) for row1, row2 in combinations(X, r=2))

	S_flattened = np.fromiter(generator, dtype=np.float64)
	S = squareform(S_flattened)
	S = lil_matrix(S)
	S.setdiag(1, k=0)
	return S.tocsr()

def _jaccard_coef_bitwise(u, v):
	u = u.astype(bool)
	v = v.astype(bool)
	return (np.double(np.bitwise_and(u,v).sum()) / np.double( np.bitwise_or(u,v).sum()))

def _jaccard_coef_set(u, v):
	u = set(u)
	v = set(v)
	inter_len = len(u.intersection(v))
	union_len = len(u) + len(v) - inter_len
	return np.double(inter_len) / np.double(union_len)

def _jaccard_coef_weighted_numpy(u,v,min_support=10):
	VMax = np.maximum(u,v) #Find maximum Values
	if np.sum(VMax) >= min_support: #Only compute when MAX()
		VMin = np.minimum(u,v) #Find minimum Values
		sumMin = np.sum(VMin) #Sum values
		sumMax = np.sum(VMax)
		coef = np.true_divide(sumMin,sumMax) # (Sum Min) / (Sum Max)
		return coef
	else:
		return 0

def _check_for_metric_type(metric):
	if metric not in __metrics__:
		raise TypeError("Metric kind should be one of: '" + "' '".join(__metrics__)+ "'")

def _get_dense_metric_function(metric):
	
	_check_for_metric_type(metric)

	if metric in ['jaccard','jb','jaccard_bitwise']:
		return _jaccard_coef_bitwise

	elif metric in ['jaccard_set','js']:
		warnings.warn('Zeros will be considered as a item set.')
		return _jaccard_coef_set

	elif metric in ['weighted_jaccard','wj']:
		return _jaccard_coef_weighted_numpy

def _get_sparse_metric_function(metric):

	_check_for_metric_type(metric)

	if metric in ['jaccard','jb','jaccard_bitwise']:
		return ('indices', _jaccard_coef_set)

	elif metric in ['jaccard_set','js']:
		return ('values', _jaccard_coef_set)

	elif metric in ['jaccard_weighted','weighted_jaccard','wj']:
		return ('both', _jaccard_coef_weighted_numpy)



# -*- coding: utf-8 -*-
"""
Compute the metric, ultrametric and semantic distance closure on the graph.

These algorithms work with undirected weighted (distance) graphs.

"""
#    Copyright (C) 2015 by
#    Luis Rocha <rocha@indiana.edu>
#    Thiago Simas <@.>
#    Rion Brattig Correia <rionbr@gmail.com>
#    All rights reserved.
#    MIT license.
import numpy as np

__author__ = """\n""".join(['Luis Rocha <rocha@indiana.com>',
							'Thiago Simas <@.>',
							'Rion Brattig Correia <rionbr@gmail.com>'])
__all__ = [
			'transitive_closure',
			'backbone_numpy',
		]

__metrics__ = ['metric','ultrametric']

def transitive_closure(A, kind='metric', *args, **kwargs):
	"""
	Compute the transitive closure (APSP) using different shortest path measures
	on the Distance Network Adjacency Matrix with values in the [0,inf] interval.
	
	- Metric: (min,+)
	- Ultrametric: (min,max)
	TODO:
	- Semantic Proximity: [ 1 + \sum_{i=2}^{n-1} log k(v_i) ]^{-1}
	
	Note: this algorithm uses numpy dense matrix calculations. Which is some slow.

		c_{ij} = min_{k}( metric ( a_{ik} , b_{kj} ) )

	Parameters
	----------
	M : Numpy matrix

    kind : string
    	type of closure to compute: 'metric', 'ultrametric' or 'semantic'

	Returns:
	------
	C : Numpy matrix
		the transitive closure graph
	
	Examples
    --------
	>>> C = transitive_closure_numpy(M, kind='metric')
	"""
	_check_for_metric_type(kind)
	
	# Numpy object
	if (type(A).__module__ == np.__name__):
		return _transitive_closure_numpy(A, kind, *args, **kwargs)
	else:
		raise TypeError("Input is not a numpy object")

def _transitive_closure_numpy(A, kind='metric', verbose=False):
	C = A.copy()
	n,m = A.shape

	# Check number of zeros in the matrix
	if len(A[A==0]) != n:
		raise ValueError("Only the diagonal has to be zero. All other zeros need to be 'numpy.inf'")
	
	# Check if diagonal is all zero
	if sum( np.diagonal(A) ) > 0:
		raise ValueError("Diagonal has to be zero for matrix computation to be correct")

	# Compute Transitive Closure
	for i in xrange(0,n):
		if verbose:
			print 'calc row:',i,'of',m-1
		for j in xrange(0,n):
			
			if kind == 'metric':
				vec = C[i,:] + C[:,j]
				C[i,j] = vec.min()

			elif kind == 'ultrametric':
				vec = np.maximum( C[i,:], C[:,j] )
				C[i,j] = vec.min()
	
	return np.array(C)


def _transitive_closure_dijkstra_matrix_all_pairs(A, kind):
	# TODO	
	return None
	n,m = A.shape

	for source in xrange(0,n):
		_transitive_closure_dijkstra_matrix(A, source, kind)

def _transitive_closure_dijkstra_matrix(A, source, kind):
	# TODO
	pass

#
# Backbone
#
def backbone(A, C):
	"""
	Return backbone edges based on the the original graph and its transitive closure.
	Values:
		1 = metric
		2 = semi-metric
		-1 = diagonal
		0 = otherwise

	Parameters
	----------
	A : dense or sparse array
		Adjacency matrix from original graph
	
	C : dense or sparse array
		Adjacency matrix from transitive closure graph

	Returns:
	------
	B : array
		Adjacency matrix where backbone edges are 1, semi-metric edges are 2, diagonal is -1, 0 otherwise.

	Examples
    --------
	>>> C = transitive_closure_numpy(A, kind='metric')
	>>> B = backbone_numpy(A, C)
	>>> rows, cols = np.where(B==1)
	"""	
	# Check for data type
	if (type(A).__module__ == np.__name__) and (type(C).__module__ == np.__name__):
		return _backbone_numpy(A, C)
	else:
		raise TypeError("Inputs are not valid objects, try numpy or scipy objects. Objects must be both the same type.")

def _backbone_numpy(A, C):
	n,m = A.shape
	B = np.zeros(A.shape)
	if n==m:
		np.fill_diagonal(B, -1)
	# Semi-metric values = 2
	rows, cols = np.where( (C > 0) & (C != np.inf) )
	B[rows,cols] = 2
	# Metric Values = 1
	rows, cols = np.where( (C > 0) & (C != np.inf) & np.isclose(A , C) )
	B[rows,cols] = 1
	return B


def _check_for_metric_type(kind):
	if kind not in __metrics__:
		raise TypeError("Metric not found for this algorithm. Try 'metric' or 'ultrametric',")




from setuptools import setup

def readme():
	with open('README.md') as f:
		return f.read()

setup(
	name='distanceclosure',
	version='0.1',
	description="Distance Closure on Complex Networks",
	long_description="Methods to calculate distance closure on complex networks as defined in T. Simas and L.M. Rocha [2015].\"Distance Closures on Complex Networks\". Network Science, 3(2):227-268. doi:10.1017/nws.2015.11",
	classifiers=[
		'Development Status :: 3 - Alpha',
		'License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)',
		'Programming Language :: Python :: 2.7',
		'Intended Audience :: Science/Research',
		'Topic :: Scientific/Engineering :: Information Analysis',
	],
	keywords="networks distance closure graph",
	url="http://github.com/rionbr/distanceclosure",
	author="Rion Brattig Correia",
	author_email="rionbr@gmail.com",
	license="GPL 2.0",
	packages=['distanceclosure'],
	install_requires=[
		'numpy',
		'scipy',
	],
	include_package_data=True,
	zip_safe=False,
	)